#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h> 

const int THR_AMT = 5;
const int MAX_TURNS = 550;
const int DEADLOCK = 0;

pthread_mutex_t mtx_lock;

/* Below, *I* is an array representing five items ("forks"): whenever a thread 
 * ("a philosopher") locks an item, we set the index of I to the thread ID. 
 * For example, if Thread 1 uses Items 1 and 2, then I[1] = 1 and I[2] = 1. 
 * The value -1 indicates that no thread has locked the item. */
char I[5] = {-1,-1,-1,-1,-1}; 

/* Here, *usage* stores the total turns of item pair usage by each thread. As per 
 * the classic problem, we assume that both adjacent items must be locked by a 
 * thread before they can be "used." */
int usage[5] = {0,0,0,0,0};   

int turn = 0;

void print_headings() {
     printf("\n────────────");
     printf("\nTurn   Items");
     printf("\n       01234\n\n");
}
 
void print_status() {

     /* What follows is a locked area: the mutex prevents race conditions with 
      * the variable *turn* and insures against the risk of corrupted output. */
     pthread_mutex_lock(&mtx_lock); 

     if (turn < MAX_TURNS) {
          if (turn % 25 == 0)
               print_headings();
          printf("%4d   ",turn);
          for (int i = 0; i < 5; i++) {
               if (I[i] != -1) {
                    printf("%d",I[i]);
               } else {
                    printf(".");
               }
          }
          printf("\n");
          turn++;
     }

     pthread_mutex_unlock(&mtx_lock);
}


void *get_resources(void *ptr) { 

     int *thread_id = (int *)ptr;

     /* Here we get the addresses of the current thread's adjacent items
      * (i.e., the "right" and "left" "forks.") */
     char *right = &I[*thread_id];
     char *left = &I[(*thread_id + 1) % 5];

     /* The following is our core logic. 
      * Set DEADLOCK to 1 to see how the classic problem results in deadlock.
      * Set DEADLOCK to 0 to see a solution using random numbers. */
     while (turn < MAX_TURNS) {

          usleep(25000); /* A philosopher "thinking" */

          if (DEADLOCK) {
               /* lock items if possible */
               if (*right == -1) 
                    *right= *thread_id;
               if (*left == -1) 
                    *left = *thread_id;
               if ((*right == *thread_id) && (*right == *left)) {
                    /* use items */
                    usleep(25000); 
                    usage[*thread_id]++;
                    /* unlock items */
                    *right = -1;
                    *left = -1;
               }         
               print_status();
          }
          
          /* Below, if a random number from 0-4 equals the current thread ID,
           * the current thread will unlock whatever items it holds, if any. 
           * Otherwise, adjacent items are locked and held, if possible. */
          if (!DEADLOCK) { 
               if (rand() % 5 == *thread_id) {
                    if (*right == *thread_id) 
                         *right = -1;
                    if (*left == *thread_id) 
                         *left = -1;
               } else {
                    if (*right == -1) 
                         *right= *thread_id;
                    if (*left == -1) 
                         *left = *thread_id;
               } 
               if ((*right == *thread_id) && (*right == *left)) {
                    print_status();
                    /* use items */
                    usleep(25000); 
                    usage[*thread_id]++;
                    /* unlock items */
                    *right = -1;
                    *left = -1;
               }
          }
          

     }
     /* A clean exit */
     pthread_exit(NULL);

} 

int main(void) { 

     int thread_return;
     int thread_arg[THR_AMT];

     pthread_t T[THR_AMT];

     for (int i = 0; i < THR_AMT; i++) {
          thread_arg[i] = i;
          thread_return = pthread_create(&T[i], NULL, get_resources, (void *) &thread_arg[i]);
          if (thread_return != 0) 
               return(thread_return); 
     }

     for (int i = 0; i < THR_AMT; i++) 
          pthread_join(T[i], NULL);  

     /* Print summary of thread usage using the variable *usage*: */
     printf("\n");
     for (int i = 0; i < THR_AMT; i++) 
          printf("Thread %d total turns of usage: %d\n", i, usage[i]);

	return 0; 
} 

